/* speakup_keyhelp.c
 * help module for speakup
 *
 *written by David Borowski.
 *
 *  Copyright (C) 2003  David Borowski.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/keyboard.h>
#include <linux/input.h>
#include "spk_priv.h"
#include "speakup.h"

#define MAXFUNCS 130
#define MAXKEYS 256
static u_short key_offsets[MAXFUNCS], key_data[MAXKEYS];

static short letter_offsets[32];

static u_char funcvals[] = {
	ATTRIB_BLEEP_DEC, ATTRIB_BLEEP_INC, BLEEPS_DEC, BLEEPS_INC,
	SAY_CHAR_ABOVE, SAY_CHAR_BELOW,
	SAY_FIRST_CHAR, SAY_LAST_CHAR, SAY_CHAR, SAY_CHAR_NUM,
	SAY_NEXT_CHAR, SAY_PHONETIC_CHAR, SAY_PREV_CHAR, TOGGLE_CURSORING,
	CURSOR_OFF, CURSOR_ON, SPEAKUP_PARKED,
	CUT, CUT1, CUT2, CUT3,
	EDIT_DELIM, EDIT_EXNUM, EDIT_MOST,
	EDIT_REPEAT, EDIT_SOME,
	FIND_PREV, FIND_NEXT, FIND_UP, FIND_DOWN,
	GOTO, TOP_EDGE, BOTTOM_EDGE,
	LEFT_EDGE, RIGHT_EDGE, SPEAKUP_HELP, CURSOR_HIGHLIGHT,
	SAY_LINE, SAY_LINE_NC, SAY_PREV_LINE, SAY_PREV_LINE_NC,
	SAY_NEXT_LINE, SAY_NEXT_LINE_NC, SAY_LINE_INDENT,
	PASTE, PASTE1, PASTE2, PASTE3,
	PITCH_DEC, PITCH_INC, PUNCT_DEC,
	PUNCT_INC, PUNC_LEVEL_DEC, PUNC_LEVEL_INC, SPEAKUP_QUIET,
	RATE_DEC, RATE_INC, READING_PUNC_DEC, READING_PUNC_INC,
	READ_ALL, READ_ALL_LINE,
	PASTE_READ, PASTE_READ1, PASTE_READ2, PASTE_READ3,
	CURSOR_WIN, ROUTE_CURSOR,
	SAY_ATTRIBUTES, SAY_FROM_LEFT, SAY_FROM_TOP,
	SAY_POSITION, SAY_SCREEN, SAY_TO_BOTTOM,
	SAY_TO_RIGHT, SET_VARS, CURSOR_SMART, SPK_KEY,
	SPK_LOCK, SPEAKUP_OFF, SPEECH_KILL,
	SPELL_DELAY_DEC, SPELL_DELAY_INC,
	SPELL_WORD, SPELL_PHONETIC,
	TOGGLE_COLUMN, TOGGLE_MULTI, TOGGLE_INDENT,
	TONE_DEC, TONE_INC, VOICE_DEC, VOICE_INC,
	VOL_DEC, VOL_INC, CLEAR_WIN, SAY_WIN,
	SET_WIN, SET_WIN_HEADER, ENABLE_WIN,
	WORD_END, SAY_WORD_ABOVE, SAY_WORD_BELOW,
	SAY_WORD, SAY_NEXT_WORD, SAY_PREV_WORD
};

static u_char *state_tbl;
static int cur_item, nstates;

static void build_key_data(void)
{
	u_char *kp, counters[MAXFUNCS], ch, ch1;
	u_short *p_key = key_data, key;
	int i, offset = 1;
	nstates = (int)(state_tbl[-1]);
	memset(counters, 0, sizeof(counters));
	memset(key_offsets, 0, sizeof(key_offsets));
	kp = state_tbl + nstates + 1;
	while (*kp++) {
		/* count occurrences of each function */
		for (i = 0; i < nstates; i++, kp++) {
			if (!*kp)
				continue;
			if (state_tbl[i] != 0 && *kp == SPK_KEY)
				continue;
			counters[*kp]++;
		}
	}
	for (i = 0; i < MAXFUNCS; i++) {
		if (counters[i] == 0)
			continue;
		key_offsets[i] = offset;
		offset += (counters[i]+1);
		if (offset >= MAXKEYS)
			break;
	}
/* leave counters set so high keycodes come first.
 * this is done so num pad and other extended keys maps are spoken before
 * the alpha with speakup type mapping.
 */
	kp = state_tbl + nstates + 1;
	while ((ch = *kp++)) {
		for (i = 0; i < nstates; i++) {
			ch1 = *kp++;
			if (!ch1)
				continue;
			if (state_tbl[i] != 0 && ch1 == SPK_KEY)
				continue;
			key = (state_tbl[i] << 8) + ch;
			counters[ch1]--;
			offset = key_offsets[ch1];
			if (!offset)
				continue;
			p_key = key_data + offset + counters[ch1];
			*p_key = key;
		}
	}
}

static void say_key(int key)
{
	int i, state = key >> 8, mask;
	char *k_name = spk_msg_get2( KEY_ID, (key&0xff));
	if (!k_name)
		return;
	if (state) {
		for (i = 5; i >= 0; i--) {
			mask = 1<<i;
			if (state & mask)
				spk_msg_say2(STATE_ID, i);
		}
	}
	synth_printf(" %s\n", k_name);
}

int say_spkup(u_char ch, u_short key)
{
	char *name = spk_msg_get2(FUNC_ID, ch);
	u_char *kp;
	int i;
	if (!name)
		return -1;
	kp = spk_our_keys[key]+1;
	for (i = 0; i < nstates-1; i++) {
		// don't use the last state as it is the catch-all for speakup key
		if (ch == kp[i])
			break;
	}
	key += (state_tbl[i] << 8);
	say_key(key);
	synth_printf(spk_msg_get(MSG_KEYDESC), name);
	synth_write("\n", 1);
	return 1;
}


static int spk_handle_help(struct vc_data *vc, u_char type, u_char ch, u_short key)
{
	int n, lastval = 0;
	char *name = NULL;
	u_char func, ch1;
	u_short *p_keys, val;
	if (type == KT_SPKUP)
		return say_spkup(ch, key);
	if (key == KEY_SPACE) {
		spk_msg_say(MSG_DONE_HELP);
		spk_special_handler = NULL;
		return 1;
	}
	if (type == KT_LATIN) {
		ch1 = ch|32; /* lower case */
		if (ch1 < 'a' || ch1 > 'z') {
			say_key(key);
			return -1;
		}
		if (letter_offsets[ch-'a'] == -1) {
			synth_printf(spk_msg_get(MSG_NO_COMMAND), ch);
			synth_buffer_add('\n');
			return 1;
		}
		cur_item = letter_offsets[ch-'a'];
	} else if (key == KEY_DOWN) {
		if (cur_item + 1 < sizeof(funcvals))
			cur_item++;
			else return -1;
	} else if (key == KEY_UP) {
		if (cur_item > 0)
			cur_item--;
		else return -1;
	} else {
		say_key(key);
		// so shift etc. doesn't keep flushing synth
		return (type == KT_SHIFT) ? 0 : 1;
	}
	func = funcvals[cur_item];
		name = spk_msg_get2(FUNC_ID, func);
	if (!name)
		return -1;
	synth_printf("%s ", name);
	if (key_offsets[func] == 0) {
		spk_msg_say(MSG_IS_UNASSIGNED);
		return 1;
	}
	p_keys = key_data + key_offsets[func];
	for (n = 0; p_keys[n]; n++) {
		val = p_keys[n];
		if (val == lastval) continue;
		if (n > 0)
			spk_msg_say_sp(MSG_DISJUNCTION);
		say_key(val);
		lastval = val;
	}
	return 1;
}

int spk_help_init(void)
{
	char start;
	int i, n, func;
	char *cur_funcname;
	for (i = 0; i < 32; i++)
		letter_offsets[i] = -1;
	state_tbl = spk_key_buf+2;
	for (i = 0; i < sizeof(funcvals); i++) {
		func = funcvals[i];
		cur_funcname = spk_msg_get2(FUNC_ID, func);
		start = *cur_funcname;
		n = (start&31)-1;
		if (letter_offsets[n] == -1)
			letter_offsets[n] = i;
	}
	cur_item = 0;
	build_key_data(); /* rebuild each time in case new mapping */
	spk_msg_say(MSG_HELP_INFO);
	spk_special_handler = spk_handle_help;
		return 1;
}
