#include <linux/slab.h> /* for kmalloc */
#include <linux/consolemap.h>
#include <linux/interrupt.h>
#include <linux/sched.h>
#include <linux/device.h> /* for dev_warn */
#include <linux/workqueue.h>
#include <linux/selection.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/atomic.h>

#include "speakup.h"

#define SPACE 0x20
#define CUTBUFFERS MAX_NR_CONSOLES+4
static char *sel_buffer[CUTBUFFERS];
static int sel_length[CUTBUFFERS];

static u_char get_char(struct vc_data *vc, int pos)
{
	u16 ch = ' ', w, c;
	w = scr_readw(screen_pos(vc, pos, 1));
	c = w & 0xff;
	if (w & vc->vc_hi_font_mask)
		c |= 0x100;
	ch = inverse_translate(vc, c, 0);
	return ch&0xff;
}


int speakup_set_selection(struct vc_data *vc, int ps, int pe, int buf_num)
{
	char *bp, *obp;
	int i, sel_start, sel_end;
	int add_cr = 0;

	if (ps > pe) {
		int tmp = ps;
		ps = pe;
		pe = tmp;
	}
	sel_start = ps;
	sel_end = pe;
	if ( pe%vc->vc_cols != 0) {
		add_cr = 1;
		while ( pe%vc->vc_cols != 0) {
			if (get_char(vc, pe++) != SPACE) {
				add_cr = 0;
				break;
			}
		}
	}

	/* Allocate a new buffer before freeing the old one ... */
	bp = kmalloc((sel_end-sel_start)+2, GFP_ATOMIC);
	if (!bp) {
		pr_warn("selection: kmalloc() failed\n");
		return -ENOMEM;
	}
	if (sel_buffer[buf_num])
		kfree(sel_buffer[buf_num]);
	sel_buffer[buf_num] = bp;

	obp = bp;
	for (i = sel_start; i <= sel_end; i++) {
		*bp = get_char(vc, i);
		if (*bp++ != ' ')
			obp = bp;
		if (!((i+1) % vc->vc_cols)) {
			// if trailing spaces on line strip them and add cr
			if (obp != bp) {
				bp = obp;
				*bp++ = '\r';
			}
			obp = bp;
		}
	}
	if (add_cr)
		*bp++ = '\r';
	sel_length[buf_num] = bp - sel_buffer[buf_num];
	return 0;
}

struct speakup_paste_work {
	struct work_struct work;
	struct tty_struct *tty;
	char *sel_buf;
	int sel_len;
};

static void __speakup_paste_selection(struct work_struct *work)
{
	struct speakup_paste_work *spw =
		container_of(work, struct speakup_paste_work, work);
	struct tty_struct *tty = xchg(&spw->tty, NULL);
	struct vc_data *vc = (struct vc_data *) tty->driver_data;
	struct tty_ldisc *ld;
	char *bp = spw->sel_buf;
	int len = spw->sel_len, count;
	DECLARE_WAITQUEUE(wait, current);

	ld = tty_ldisc_ref(tty);
	if (!ld)
		goto tty_unref;
	tty_buffer_lock_exclusive(&vc->port);

	add_wait_queue(&vc->paste_wait, &wait);
	while (len > 0) {
		set_current_state(TASK_INTERRUPTIBLE);
		if (test_bit(TTY_THROTTLED, &tty->flags)) {
			schedule();
			continue;
		}
		count = len;
		count = tty_ldisc_receive_buf(ld, bp, NULL, count);
		len -= count;
		bp += count;
	}
	remove_wait_queue(&vc->paste_wait, &wait);
	__set_current_state(TASK_RUNNING);

	tty_buffer_unlock_exclusive(&vc->port);
	tty_ldisc_deref(ld);
tty_unref:
	tty_kref_put(tty);
}

static struct speakup_paste_work speakup_paste_work = {
	.work = __WORK_INITIALIZER(speakup_paste_work.work,
				   __speakup_paste_selection)
};

int speakup_paste_selection(struct tty_struct *tty, int buf_num)
{
	if (!sel_buffer[buf_num]) return -1;
	if (cmpxchg(&speakup_paste_work.tty, NULL, tty) != NULL)
		return -EBUSY;

	speakup_paste_work.sel_buf = sel_buffer[buf_num];
	speakup_paste_work.sel_len = sel_length[buf_num];
	tty_kref_get(tty);
	schedule_work_on(WORK_CPU_UNBOUND, &speakup_paste_work.work);
	return 0;
}

void speakup_init_selection( void ) {
	memset( sel_buffer, 0, sizeof( sel_buffer ) );
	memset( sel_length, 0, sizeof( sel_length ) );
}

void speakup_free_selection( void ) {
	int i;
	cancel_work_sync(&speakup_paste_work.work);
	tty_kref_put(speakup_paste_work.tty);
	for ( i = 0; i < CUTBUFFERS; i++ ) {
		if (sel_buffer[i])
			kfree(sel_buffer[i]);
	}
}

int speakup_read_selection(int buf_num)
{
	int     len = sel_length[buf_num];
	char *bp = sel_buffer[buf_num];
	if (bp && len > 0) {
		spkup_write(bp, len, spk_reading_mask);
		synth_buffer_add('\n');
	} else spk_msg_say(MSG_BLANK);
	return 0;
}
